@@ -1,6 +1,6 @@
 #!/usr/bin/env python3
 # thoth-python
-# Copyright(C) 2018, 2019 Fridolin Pokorny
+# Copyright(C) 2018, 2019, 2020 Fridolin Pokorny
 #
 # This program is free software: you can redistribute it and / or modify
 # it under the terms of the GNU General Public License as published by
@@ -17,6 +17,7 @@
 
 """Project abstraction and operations on project dependencies."""
 
+import os
 import logging
 import typing
 from itertools import chain
@@ -38,6 +39,8 @@ from .package_version import PackageVersion
 from .exceptions import UnableLock
 from .exceptions import InternalError
 from .exceptions import NotFound
+from .exceptions import FileLoadError
+from .helpers import parse_requirements
 
 _LOGGER = logging.getLogger(__name__)
 
@@ -71,13 +74,28 @@ class Project:
         without_pipfile_lock: bool = False,
     ):
         """Create project from Pipfile and Pipfile.lock files."""
-        with open(pipfile_path or "Pipfile", "r") as pipfile_file:
-            pipfile = Pipfile.from_string(pipfile_file.read())
+        try:
+            with open(pipfile_path or "Pipfile", "r") as pipfile_file:
+                pipfile_str = pipfile_file.read()
+        except Exception as exc:
+            raise FileLoadError(
+                f"Failed to load Pipfile (path: {os.getcwd() if not pipfile_path else pipfile_path}: {str(exc)}"
+            ) from exc
+
+        pipfile = Pipfile.from_string(pipfile_str)
 
         pipfile_lock = None
         if not without_pipfile_lock:
-            with open(pipfile_lock_path or "Pipfile.lock", "r") as pipfile_lock_file:
-                pipfile_lock = PipfileLock.from_string(pipfile_lock_file.read(), pipfile)
+            try:
+                with open(pipfile_lock_path or "Pipfile.lock", "r") as pipfile_lock_file:
+                    pipfile_lock_str = pipfile_lock_file.read()
+            except Exception as exc:
+                raise FileLoadError(
+                    f"Failed to load Pipfile.lock "
+                    f"(path: {os.getcwd() if not pipfile_lock_path else pipfile_lock_path}: {str(exc)}"
+                ) from exc
+
+            pipfile_lock = PipfileLock.from_string(pipfile_lock_str, pipfile=pipfile)
 
         return cls(
             pipfile,
@@ -85,6 +103,21 @@ class Project:
             runtime_environment=runtime_environment if runtime_environment else RuntimeEnvironment.from_dict({})
         )
 
+    @classmethod
+    def from_dict(
+        cls,
+        pipfile: typing.Dict[str, typing.Any],
+        pipfile_lock: typing.Dict[str, typing.Any],
+        runtime_environment: RuntimeEnvironment = None
+    ) -> "Project":
+        """Construct project out of a dict representation."""
+        pipfile = Pipfile.from_dict(pipfile)
+        return cls(
+            pipfile=pipfile,
+            pipfile_lock=PipfileLock.from_dict(pipfile_lock, pipfile=pipfile),
+            runtime_environment=runtime_environment
+        )
+
     @classmethod
     def from_strings(
         cls, pipfile_str: str, pipfile_lock_str: str = None, *, runtime_environment: RuntimeEnvironment = None
@@ -102,7 +135,12 @@ class Project:
             runtime_environment=runtime_environment if runtime_environment else RuntimeEnvironment.from_dict({})
         )
 
-    def to_files(self, pipfile_path: str = None, pipfile_lock_path: str = None, without_pipfile_lock: bool = False):
+    def to_files(
+        self,
+        pipfile_path: str = None,
+        pipfile_lock_path: str = None,
+        without_pipfile_lock: bool = False
+    ) -> None:
         """Write the current state of project into Pipfile and Pipfile.lock files."""
         with open(pipfile_path or "Pipfile", "w") as pipfile_file:
             pipfile_file.write(self.pipfile.to_string())
@@ -111,12 +149,106 @@ class Project:
             with open(pipfile_lock_path or "Pipfile.lock", "w") as pipfile_lock_file:
                 pipfile_lock_file.write(self.pipfile_lock.to_string())
 
+    def construct_requirements_in(self) -> str:
+        """Construct requirements.in file for the current project."""
+        return self.pipfile.construct_requirements_in()
+
+    def construct_requirements_txt(self) -> str:
+        """Construct requirements.txt file for the current project - pip-tools compatible."""
+        return self.pipfile_lock.construct_requirements_txt()
+
+    def to_pip_compile_files(
+        self,
+        requirements_path: str = "requirements.in",
+        requirements_lock_path: str = "requirements.txt",
+        without_lock: bool = False
+    ) -> None:
+        """Write the current state of project into requirements.in and requirements.txt files.
+
+        Files created are compatible with pip/pip-tools. If no lock is specified, requirements.txt
+        hold unpinned direct dependencies as in case of pip.
+        """
+        if without_lock:
+            with open(requirements_path, "w") as requirements_file:
+                requirements_file.write(self.construct_requirements_in())
+            return
+
+        with open(requirements_path, "w") as requirements_file:
+            requirements_file.write(self.construct_requirements_in())
+
+        with open(requirements_lock_path, "w") as requirements_lock_file:
+            requirements_lock_file.write(self.construct_requirements_txt())
+
+    @classmethod
+    def from_pip_compile_files(
+        cls,
+        requirements_path: str = "requirements.in",
+        requirements_lock_path: typing.Optional[str] = None,
+        allow_without_lock: bool = False,
+        runtime_environment: typing.Optional[RuntimeEnvironment] = None,
+    ) -> "Project":
+        """Parse project from files compatible with pip/pip-tools."""
+        sources_lock, package_versions_lock = None, None
+        if allow_without_lock:
+            if not os.path.exists(requirements_path):
+                requirements_lock_path = requirements_lock_path or "requirements.txt"
+                sources, package_versions = parse_requirements(requirements_lock_path)
+            else:
+                sources, package_versions = parse_requirements(requirements_path)
+                if requirements_lock_path is not None:
+                    sources_lock, package_versions_lock = parse_requirements(requirements_lock_path)
+                elif os.path.exists("requirements.txt"):
+                    sources_lock, package_versions_lock = parse_requirements("requirements.txt")
+        else:
+            requirements_lock_path = requirements_lock_path or "requirements.txt"
+            sources, package_versions = parse_requirements(requirements_path)
+            sources_lock, package_versions_lock = parse_requirements(requirements_lock_path)
+
+        for package_version in package_versions_lock or []:
+            if not package_version.is_locked():
+                raise ValueError(
+                    f"Package {package_version.name} in lockfile does not "
+                    f"use locked version: {package_version.version!r}"
+                )
+
+            if not package_version.hashes:
+                _LOGGER.warning(
+                    "Parsed package %r in version %r from lockfile %r has no hashes assigned",
+                    package_version.name,
+                    package_version.version,
+                    requirements_lock_path,
+                )
+
+        sources = sources or sources_lock or []
+
+        # We know which index is assigned.
+        if len(sources) == 1:
+            for package_version in package_versions:
+                package_version.index = sources[0]
+            for package_version in package_versions_lock or []:
+                package_version.index = sources[0]
+
+        # XXX: do we want a default source?
+        meta = PipfileMeta(sources={s.name: s for s in sources})
+        pipfile = Pipfile.from_package_versions(package_versions, meta=meta)
+        pipfile_lock = None
+        if package_versions_lock:
+            pipfile_lock = PipfileLock.from_package_versions(
+                pipfile=pipfile,
+                packages=package_versions_lock,
+                meta=meta
+            )
+
+        return cls(pipfile, pipfile_lock, runtime_environment=runtime_environment)
+
     @classmethod
     def from_package_versions(
         cls,
         packages: typing.List[PackageVersion],
         packages_locked: typing.List[PackageVersion] = None,
         meta: PipfileMeta = None,
+        *,
+        runtime_environment: RuntimeEnvironment = None,
     ):
         """Create project from PackageVersion objects.
 
@@ -126,9 +258,14 @@ class Project:
         pipfile = Pipfile.from_package_versions(packages, meta=meta)
         pipfile_lock = None
         if packages_locked:
-            pipfile_lock = PipfileLock.from_package_versions(pipfile, packages_locked, meta=meta)
+            pipfile_lock = PipfileLock.from_package_versions(pipfile, packages_locked, meta=pipfile.meta)
+
+        if runtime_environment:
+            instance = cls(pipfile, pipfile_lock, runtime_environment=runtime_environment)
+        else:
+            # Let default be expanded.
+            instance = cls(pipfile, pipfile_lock)
 
-        instance = cls(pipfile, pipfile_lock)
         instance.sanitize_source_indexes()
         return instance
 
