@@ -1,6 +1,6 @@
 #!/usr/bin/env python3
 # thoth-python
-# Copyright(C) 2018, 2019 Fridolin Pokorny
+# Copyright(C) 2018, 2019, 2020 Fridolin Pokorny
 #
 # This program is free software: you can redistribute it and / or modify
 # it under the terms of the GNU General Public License as published by
@@ -20,19 +20,21 @@
 import logging
 import re
 import typing
-import hashlib
 from functools import lru_cache
 from urllib.parse import urlparse
 
 import attr
 import requests
 from bs4 import BeautifulSoup
-import semantic_version as semver
+from packaging.version import Version
+from packaging.version import LegacyVersion
+from packaging.version import parse as parse_version
 
 from .exceptions import NotFound
 from .exceptions import InternalError
 from .exceptions import VersionIdentifierError
 from .configuration import config
+from .artifact import Artifact
 
 _LOGGER = logging.getLogger(__name__)
 
@@ -125,13 +127,22 @@ class Source:
         response.raise_for_status()
         return response.json()
 
-    def _warehouse_get_package_hashes(self, package_name: str, package_version: str) -> typing.List[dict]:
+    def _warehouse_get_package_hashes(
+        self, package_name: str, package_version: str, with_included_files: bool = False
+    ) -> typing.List[dict]:
         """Gather information about SHA hashes available for the given package-version release."""
         package_info = self._warehouse_get_api_package_version_info(package_name, package_version)
 
         result = []
         for item in package_info["urls"]:
-            result.append({"name": item["filename"], "sha256": item["digests"]["sha256"]})
+            result.append(
+                {"name": item["filename"], "sha256": item["digests"]["sha256"]}
+            )
+            # this checks whether to gather digests for all files in the given artifact
+            if with_included_files:
+                artifact = Artifact(item["filename"], item["url"], verify_ssl=self.verify_ssl)
+                result[-1]["digests"] = artifact.gather_hashes()
+                result[-1]["symbols"] = artifact.get_versioned_symbols()
 
         return result
 
@@ -165,11 +176,28 @@ class Source:
                 #   https://www.python.org/dev/peps/pep-0503/
                 link_text = link_text[:-1]
 
+            # Normalize link_text for comparison below
+            link_text = self.normalize_package_name(link_text)
+
             if self.is_normalized_python_package_name(package_name) and package_name == link_text:
                 packages.add(package_name)
 
         return packages
 
+    def provides_package(self, package_name: str) -> bool:
+        """Check if the given package is provided by this package source index."""
+        _LOGGER.debug("Checking availability of package %r on index %r", package_name, self.url)
+        package_name = self.normalize_package_name(package_name)
+        url = self.url + "/" + package_name
+        response = requests.get(url, verify=self.verify_ssl)
+
+        if response.status_code == 404:
+            return False
+
+        # Raise on other inappropriate error codes.
+        response.raise_for_status()
+        return True
+
     @staticmethod
     def _parse_artifact_version(package_name: str, artifact_name: str) -> str:
         """Parse package version based on artifact name available on the package source index."""
@@ -215,7 +243,7 @@ class Source:
 
     def get_latest_package_version(
         self, package_name: str, graceful: bool = False
-    ) -> typing.Optional[semver.base.Version]:
+    ) -> typing.Optional[typing.Union[Version, LegacyVersion]]:
         """Get the latest version for the given package."""
         try:
             all_versions = self.get_package_versions(package_name)
@@ -229,7 +257,7 @@ class Source:
         semver_versions = []
         for version in all_versions:
             try:
-                version = semver.Version.coerce(version)
+                version = parse_version(version)
             except Exception as exc:
                 error_msg = f"Cannot parse semver version {version} for package {package_name}: {str(exc)}"
                 if graceful:
@@ -278,15 +306,18 @@ class Source:
 
         return artifacts
 
-    def _download_artifacts_sha(self, package_name: str, package_version: str) -> typing.Generator[tuple, None, None]:
-        """Download the given artifact from Warehouse and compute its SHA."""
+    def get_package_artifacts(self, package_name: str, package_version: str):
+        """Return list of artifacts corresponding to package name and package version."""
+        to_return = []
+        possible_continuations = [".win32", ".tar.gz", ".whl", ".zip", ".exe", ".egg", "-"]
         for artifact_name, artifact_url in self._simple_repository_list_artifacts(package_name):
             # Convert all artifact names to lowercase - as a shortcut we simply convert everything to lowercase.
             artifact_name.lower()
-            if not artifact_name.startswith(f"{package_name}-{package_version}"):
-                # TODO: this logic has to be improved as package version can be a suffix of another package version:
-                #   mypackage-1.0.whl, mypackage-1.0.0.whl, ...
-                # This will require parsing based on PEP or some better logic.
+
+            for i in possible_continuations:
+                if artifact_name.startswith(f"{package_name}-{package_version}{i}"):
+                    break
+            else:
                 _LOGGER.debug(
                     "Skipping artifact %r as it does not match required version %r for package %r",
                     artifact_name,
@@ -294,25 +325,41 @@ class Source:
                     package_name,
                 )
                 continue
+            to_return.append(Artifact(artifact_name, artifact_url, verify_ssl=self.verify_ssl))
 
-            url_parts = artifact_url.rsplit("#", maxsplit=1)
-            if len(url_parts) == 2 and url_parts[1].startswith("sha256="):
-                _LOGGER.debug("Using SHA256 stated in URL: %r", url_parts[1])
-                yield artifact_name, url_parts[1][len("sha256="):]
+        return to_return
+
+    def _download_artifacts_data(
+        self, package_name: str, package_version: str, with_included_files: bool = False
+    ) -> typing.Generator[tuple, None, None]:
+        """Download the given artifact from Warehouse and compute desired info."""
+        possible_continuations = [".win32", ".tar.gz", ".whl", ".zip", ".exe", ".egg", "-"]
+        for artifact_name, artifact_url in self._simple_repository_list_artifacts(package_name):
+            # Convert all artifact names to lowercase - as a shortcut we simply convert everything to lowercase.
+            artifact_name.lower()
+
+            for i in possible_continuations:
+                if artifact_name.startswith(f"{package_name}-{package_version}{i}"):
+                    break
+            else:
+                _LOGGER.debug(
+                    "Skipping artifact %r as it does not match required version %r for package %r",
+                    artifact_name,
+                    package_version,
+                    package_name,
+                )
                 continue
+            artifact = Artifact(artifact_name, artifact_url, verify_ssl=self.verify_ssl)
 
-            _LOGGER.debug("Downloading artifact from url %r", artifact_url)
-            response = requests.get(artifact_url, verify=self.verify_ssl, stream=True)
-            response.raise_for_status()
+            symbols = None
+            hashes = None
+            if with_included_files:
+                symbols = artifact.get_versioned_symbols()
+                hashes = artifact.gather_hashes()
 
-            digest = hashlib.sha256()
-            for chunk in response.iter_content(chunk_size=1024):
-                if chunk:  # filter out keep-alive new chunks
-                    digest.update(chunk)
-
-            digest = digest.hexdigest()
-            _LOGGER.debug("Computed artifact sha256 digest for %r: %s", artifact_url, digest)
-            yield artifact_name, digest
+            yield (
+                artifact_name, artifact.sha, hashes, symbols,
+            )
 
     def provides_package_version(self, package_name: str, package_version: str) -> bool:
         """Check if the given source provides package in the given version."""
@@ -323,12 +370,38 @@ class Source:
             return False
 
     @lru_cache(maxsize=10)
-    def get_package_hashes(self, package_name: str, package_version: str) -> list:
+    def get_package_data(self, package_name: str, package_version: str, with_included_files: bool = False) -> list:
+        """Get information about release hashes and symbols available in this source index."""
+        if self.warehouse:
+            return self._warehouse_get_package_hashes(package_name, package_version, with_included_files)
+
+        artifacts = self.get_package_artifacts(package_name, package_version)
+        result = []
+        for artifact in artifacts:
+            doc = {}
+            doc["name"] = artifact.artifact_name
+            doc["sha256"] = artifact.sha
+            if with_included_files:
+                doc["digests"] = artifact.gather_hashes()
+                doc["symbols"] = artifact.get_versioned_symbols()
+            result.append(doc)
+
+        return result
+
+    @lru_cache(maxsize=10)
+    def get_package_hashes(self, package_name: str, package_version: str, with_included_files: bool = False) -> list:
         """Get information about release hashes available in this source index."""
         if self.warehouse:
-            return self._warehouse_get_package_hashes(package_name, package_version)
+            return self._warehouse_get_package_hashes(package_name, package_version, with_included_files)
 
-        return [
-            {"name": name, "sha256": digest}
-            for name, digest in self._download_artifacts_sha(package_name, package_version)
-        ]
+        artifacts = self.get_package_artifacts(package_name, package_version)
+        result = []
+        for artifact in artifacts:
+            doc = {}
+            doc["name"] = artifact.artifact_name
+            doc["sha256"] = artifact.sha
+            if with_included_files:
+                doc["digests"] = artifact.gather_hashes()
+            result.append(doc)
+
+        return result
